﻿using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;

using System;
using System.Collections.Generic;

using System.Linq;

using System.Threading;
using System.Reflection;
using System.Diagnostics;
using System.IO;

/////////////////////////////////////////////////////////////////////////////////
/// <summary>
/// 連続バルーン
/// 概要
/// 　自動で数字や文字をカウントアップしながら連続してバルーン図形を作成する RootPro CAD のアドインプログラムです。
/// 　既存のバルーンを選択することで連続した文字に置き換えることもできます。
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
///
/// Copyright @ 2023 RootPro Co.,Ltd. (Japan)
/// </summary>
/// <summary locale = "en">
/// Continuous Balloon
/// overview
/// This is an add-in program for RootPro CAD that automatically counts up numbers and letters and creates balloon shape continuously.
/// You can also select an existing balloon and replace it with consecutive characters.
/// License
///  The user is solely responsible for the result of executing this add-in.
///  This add-in can modify or divert the program source itself.
///  We also allow you to create your own add-ins and software based on this program source and distribute them to third parties regardless of whether they are free or paid.
/// 
/// Copyright @ 2023 RootPro Co., Ltd. (Japan)
/// </summary>

namespace RCAddInContinuousBalloonCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
         // 設定値
        AddInSettings addInSettings = null;

        string CommandName_A = "RootPro.AddIn.ContinuousBalloonCommand";     //コマンド名 

        // コマンドパラメータID
        const int countupModeParameterId = 1;               // カウントアップモードパラメータID
        const int startBalloonShapeParameterId = 2;         // カウントを始めるバルーン図形パラメータID
        const int textParameterId = 3;                      // 文字パラメータID
        const int startPointParameterId = 4;                // 始点パラメータID
        const int endPointParameterId = 5;                  // 終点パラメータID
        const int pointListParameterId = 6;                 // 点列パラメータID
        const int locatePointParameterId = 7;               // 配置点パラメータID
        const int replaceBalloonParameterId = 8;            // 置き換えるバルーン図形パラメータID
        const int replaceSameTextBalloonFlgParameterId = 9; // 同じ文字のバルーンも置き換えるフラグパラメータID

        /// <summary>
        /// カウントアップモード
        /// </summary>
        enum CountupMode
        {
            CountupMode_Auto = 0,      // 自動
            CountupMode_NextSelectedBalloon = 1, // 選択したバルーンの次から
            CountupMode_Manual = 2,// 手動
        }

         // エラー番号
        const int RC_ERR_SUCCESS = 1; // エラーなし（成功）
        const int RC_ERR_NONPOINTS = -1; // エラーなし（点列なし）

        bool changedTextParameter = false;// 文字パラメーターを自動的に変更したフラグ

        // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」パラメーターグループ
        CommandParameterGroup locateBalloonModeParameterGroup = null;

        // バルーンに使う文字のリスト(数字以外)
        string[] balloonTextList = {
        "abcdefghijklmnopqrstuvwxyz",
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
        "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ",
        "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ",
        "アイウエオカキクケコサシスセソタチツテトナニヌネノハヒフヘホマミムメモヤユヨラリルレロワン",
        "あいうえおかきくけこさしすせそたちつてとなにぬねのはひふへほまみむめもやゆよらりるれろわん"};

        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // ファイルから設定値を読み込む
            if(!LoadAddInSettings(out addInSettings))
                addInSettings = new AddInSettings();// 読み込めなかった場合は初期値

            // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

			CommandManager.AddMacroCommand("RCAddInContinuousBalloonCS", Properties.Resources.CommandName, MacroCommand);// リソースファイルを追加してコマンド名などを定義しておく

            // 図形コマンドを追加する
            Command command = CommandManager.AddCommand(CommandName_A, Properties.Resources.CommandName);// AddIn のコマンド名をセット
	        AddParameters(command);// コマンドパラメーターセット

            // 開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            // 設定値をファイルに書き込む
            Debug.Assert(addInSettings != null);
            SaveAddInSettings(addInSettings);

            // コマンドの削除
            CommandManager.RemoveMacroCommand(MacroCommand);
            CommandManager.RemoveCommand(CommandName_A);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 外部フォルダのアセンブリを動的にロード
        /// 多言語対応リソースのDLLファイルをロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
		{
			AssemblyName assemblyName = new AssemblyName(args.Name);

			foreach (Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
			{
				if (AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
					return assembly;
			}

			if (assemblyName.Name.EndsWith(".resources"))
			{
				if (assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
				{
					string dllPath = Path.GetDirectoryName(this.AddInFullPath);
					dllPath = Path.Combine(dllPath, assemblyName.CultureName);
					dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

					if (File.Exists(dllPath))
					{
						try
						{
							return Assembly.LoadFrom(dllPath);
						}
						catch
						{
						}
					}
				}
			}

			return null;
		}

        private void MacroCommand()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(CommandName_A);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            // パラメータアイテムの変更イベントの追加
            command.ParameterGroupCurrentIndexChanged += new CommandParameterGroupCurrentIndexChangedEventHandler(command_ParameterGroupCurrentIndexChanged);

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);

            //マウス移動イベントの追加
            command.MouseMove += new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            // パラメータアイテムの変更イベントの削除
            command.ParameterGroupCurrentIndexChanged -= new CommandParameterGroupCurrentIndexChangedEventHandler(command_ParameterGroupCurrentIndexChanged);

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);

            //マウス移動イベントの削除
            command.MouseMove -= new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// パラメータを追加する
        /// </summary>
        /// <param name="command"></param>
        /// <param name="InputParameterType"></param>
        private void AddParameters(Command command)
        {
            // コマンドウィンドウの各行にパラメータを追加
            // メモ：コマンドウィンドウでコマンドのパラメーターを用意する
            // 各パラメーターは ParameterItemType がありそれぞれ扱いたい値によって使い分ける
            // パラメータの種類
            // ParameterItemType.Point      --- PointParameterItem;      // 座標パラメータ         -- 座標値を入力できます。マウスで図面上から指定することも可能です。
            // ParameterItemType.PointList  --- PointListParameterItem;  // 点列パラメータ         -- 連続した複数の座標値を入力できます。
            // ParameterItemType.Length     --- LengthParameterItem;     // 長さパラメータ         -- 長さを入力できます。
            // ParameterItemType.Angle      --- AngleParameterItem;      // 角度パラメータ         -- 角度を入力できます。
            // ParameterItemType.Double     --- DoubleParameterItem;     // 倍精度浮動小数点数パラメータ
            // ParameterItemType.Integer    --- IntegerParameterItem;    // 整数パラメータ
            // ParameterItemType.Scale      --- ScaleParameterItem;      // 倍率パラメータ
            // ParameterItemType.Text       --- TextParameterItem;       // 文字パラメータ         -- 文字列を入力できます。
            // ParameterItemType.List       --- ListParameterItem;       // リストパラメータ       -- 選択肢をセットして選ぶことができます。
            // ParameterItemType.Shape      --- ShapeParameterItem;      // 図形パラメータ         -- 図形が選択できます。
            // ParameterItemType.ShapeGroup --- ShapeGroupParameterItem; // 図形グループパラメータ -- 複数の図形が選択できます。
            // ParameterItemType.Multiple   --- MultipleParameterItem;   // 複合パラメータ

            // パラメータを追加する
            // 「モード」
            CommandParameterGroup commandParameterGroup = command.ParameterGroups.Add(ParameterItemType.List, countupModeParameterId, Properties.Resources.CountupModeParameterName);// 1つ目のパラメータを追加するとパラメータグループが作られ、その中のItemの一つかここで追加したパラメータになる
            ListParameterItem modeListParameterItem = (ListParameterItem)command.GetParameterItemByID(countupModeParameterId);// 追加したパラメータのItemを取得
            string itemsString = Properties.Resources.CountupModeListItems;
            string[] items = itemsString.Split(',');
            modeListParameterItem.Items = items;
            modeListParameterItem.SelectedIndex = addInSettings.countupModeListParameterItemIndex;
            modeListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            modeListParameterItem.Description = Properties.Resources.CountupModeListParameterDescription;
            // 「モード」が「選択したバルーンの次から」の場合
            if(modeListParameterItem.SelectedIndex == (int)CountupMode.CountupMode_NextSelectedBalloon)
            {
                // 「カウントを始めるバルーン図形」パラメーターを追加
                command.ParameterGroups.Add(ParameterItemType.Shape, startBalloonShapeParameterId, Properties.Resources.StartBalloonShapeParameterName);
                ShapeParameterItem shapeParameterItem = (ShapeParameterItem)command.GetParameterItemByID(startBalloonShapeParameterId);
                shapeParameterItem.ShapeTypeFilter = ShapeTypeFilters.Balloon;
                shapeParameterItem.Description = Properties.Resources.StartBalloonShapeParameterDescription;
                shapeParameterItem.ValueSaveMode = ParameterValueSaveMode.InCommand;// コマンド実行中は保持
            }
           
            // 「文字」
            command.ParameterGroups.Add(ParameterItemType.Text, textParameterId, Properties.Resources.TextParameterName);
            TextParameterItem textParameterItem = (TextParameterItem)command.GetParameterItemByID(textParameterId); // パラメータアイテム
            textParameterItem.Description = Properties.Resources.TextParameterDescription;
            if(modeListParameterItem.SelectedIndex == (int)CountupMode.CountupMode_Manual)
                textParameterItem.ValueSaveMode = ParameterValueSaveMode.OneTime;
            else
                textParameterItem.ValueSaveMode = ParameterValueSaveMode.Always;
            textParameterItem.Text = addInSettings.textParameterValue;// 設定の値を初期値としてセット

            // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」をグループ登録する(パラメーター名の左右に矢印が表示され切り替えができるようになる)
            // 「始点」
            locateBalloonModeParameterGroup = command.ParameterGroups.Add(ParameterItemType.Point, startPointParameterId, Properties.Resources.StartPointParameterName);
            PointParameterItem startPointParameterItem = (PointParameterItem)command.GetParameterItemByID(startPointParameterId);
            startPointParameterItem.Description = Properties.Resources.StartPointParameterDescription;
            if(addInSettings.locateBalloonModeItemID == startPointParameterId)
            {
                locateBalloonModeParameterGroup.CurrentItem = startPointParameterItem;

                // 「終点」
                command.ParameterGroups.Add(ParameterItemType.Point, endPointParameterId, Properties.Resources.EndPointParameterName);
                PointParameterItem endPointParameterItem = (PointParameterItem)command.GetParameterItemByID(endPointParameterId);
                endPointParameterItem.Description = Properties.Resources.EndPointParameterDescription;
            }

            // 「点列」
            locateBalloonModeParameterGroup.Items.Add(ParameterItemType.PointList, pointListParameterId, Properties.Resources.PointListParameterName);
            PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetParameterItemByID(pointListParameterId);
            pointListParameterItem.Description = Properties.Resources.PointListParameterDescription;
            if(addInSettings.locateBalloonModeItemID == pointListParameterId)
                locateBalloonModeParameterGroup.CurrentItem = pointListParameterItem;

            // 「配置点」
            locateBalloonModeParameterGroup.Items.Add(ParameterItemType.Point, locatePointParameterId, Properties.Resources.LocatePointParameterName);
            PointParameterItem locatePointParameterItem = (PointParameterItem)command.GetParameterItemByID(locatePointParameterId);
            locatePointParameterItem.Description = Properties.Resources.LocatePointParameterDescription;
            if(addInSettings.locateBalloonModeItemID == locatePointParameterId)
                locateBalloonModeParameterGroup.CurrentItem = locatePointParameterItem;

            // 「置き換えるバルーン図形」
            locateBalloonModeParameterGroup.Items.Add(ParameterItemType.Shape, replaceBalloonParameterId, Properties.Resources.ReplaceBalloonParameterName);
            ShapeParameterItem replaceBalloonParameterItem = (ShapeParameterItem)command.GetParameterItemByID(replaceBalloonParameterId);
            replaceBalloonParameterItem.Description = Properties.Resources.ReplaceBalloonParameterDescription;
            replaceBalloonParameterItem.ShapeTypeFilter = ShapeTypeFilters.Balloon;// バルーンのみ
            replaceBalloonParameterItem.EditableShapeOnly = true;// 変更可能な図形のみ選択可能
            if(addInSettings.locateBalloonModeItemID == replaceBalloonParameterId)
            {
                locateBalloonModeParameterGroup.CurrentItem = replaceBalloonParameterItem;

                //「同じ文字のバルーンも置き換える」を追加
                command.ParameterGroups.Add(ParameterItemType.List, replaceSameTextBalloonFlgParameterId, Properties.Resources.ReplaceSameTextBalloonFlgParameterName);
                ListParameterItem modeListParameterItem2 = (ListParameterItem)command.GetParameterItemByID(replaceSameTextBalloonFlgParameterId);
                string itemsString2 = Properties.Resources.ReplaceSameTextBalloonFlgItems;
                string[] items2 = itemsString2.Split(',');
                modeListParameterItem2.Items = items2;
                modeListParameterItem2.SelectedIndex = addInSettings.replaceSameTextBalloonFlgIndex;
                modeListParameterItem2.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
                modeListParameterItem2.Description = Properties.Resources.ReplaceSameTextBalloonFlgParameterDescription;
            }
           
        }

        /// <summary>
        /// パラメータグループ(パラメータ名の左右に矢印がある状態)のカレントのパラメータアイテムが変更されたときに呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_ParameterGroupCurrentIndexChanged(Object sender, CommandParameterGroupCurrentIndexChangedEventArgs e)
        {

            Document doc = ActiveDocument;         //ドキュメントの取得

            Command command = (Command)sender;

            CommandParameterGroup parameterGroup = e.ParameterGroup;    // 変更されたパラメータグループ

            // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」パラメーターグループが変わったとき
            Debug.Assert(locateBalloonModeParameterGroup != null);
            if(locateBalloonModeParameterGroup.Equals(parameterGroup))
            {
                // 変わる前後のパラメータID
                ParameterItemCollection items = parameterGroup.Items;
                ParameterItem oldItem = items[e.OldIndex];
                ParameterItem newItem = items[e.NewIndex];
                int oldItemID = oldItem.ID;
                int newItemID = newItem.ID;

                // 「始点」→ 他のパラメーター
                if(oldItemID == startPointParameterId)
                {
                    // 「終点」を削除
                    ParameterItem parameterItem = command.GetParameterItemByID(endPointParameterId);
                    if(parameterItem != null)
                    {
                        CommandParameterGroup parameterItemGroup = parameterItem.Group;
                        command.ParameterGroups.Remove(parameterItemGroup);
                    }
                }

                // 他のパラメーター →「始点」
                if(newItemID == startPointParameterId)
                {
                    // 「終点」を最後に追加
                    command.ParameterGroups.Add(ParameterItemType.Point, endPointParameterId, Properties.Resources.EndPointParameterName);
                    PointParameterItem endPointParameterItem = (PointParameterItem)command.GetParameterItemByID(endPointParameterId);
                    endPointParameterItem.Description = Properties.Resources.EndPointParameterDescription;
                }

                //「同じ文字のバルーンも置き換える」→ 他のパラメーター
                if(oldItemID == replaceBalloonParameterId)
                {
                    // 「同じ文字のバルーンも置き換える」を削除
                    ParameterItem parameterItem = command.GetParameterItemByID(replaceSameTextBalloonFlgParameterId);
                    if(parameterItem != null)
                    {
                        CommandParameterGroup parameterItemGroup = parameterItem.Group;
                        command.ParameterGroups.Remove(parameterItemGroup);
                    }
                }

                // 他のパラメーター →「同じ文字のバルーンも置き換える」
                if(newItemID == replaceBalloonParameterId)
                {
                    //「同じ文字のバルーンも置き換える」を追加
                    command.ParameterGroups.Add(ParameterItemType.List, replaceSameTextBalloonFlgParameterId, Properties.Resources.ReplaceSameTextBalloonFlgParameterName);
                    ListParameterItem modeListParameterItem2 = (ListParameterItem)command.GetParameterItemByID(replaceSameTextBalloonFlgParameterId);
                    string itemsString2 = Properties.Resources.ReplaceSameTextBalloonFlgItems;
                    string[] items2 = itemsString2.Split(',');
                    modeListParameterItem2.Items = items2;
                    modeListParameterItem2.SelectedIndex = addInSettings.replaceSameTextBalloonFlgIndex;
                    modeListParameterItem2.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
                    modeListParameterItem2.Description = Properties.Resources.ReplaceSameTextBalloonFlgParameterDescription;
                }
            }
        }


        /// <summary>
        /// パラメータの変化のイベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Command command = (Command)sender;

            // 文字パラメーターを自動的に変更した直後の場合は何もしない
            // 「文字」パラメーターに次の文字をセットしたときにこのイベントが発生する
            if(changedTextParameter)
            {
                changedTextParameter = false;
                return;
            }
            
            // 「モード」のパラメータアイテム取得
            ListParameterItem countupModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(countupModeParameterId);
            if(countupModeListParameterItem == null)
                return;

            TextParameterItem textParameterItem = null;

            // 選ばれているモードのインデックス
            int countupModeSelectedIndex = countupModeListParameterItem.SelectedIndex;

            int changedParameterId = e.ParameterID;// 変更のあったパラメータID
            // 「モード」切り替え時(▼ボタンのあるコンボボックスで選ばれている内容が変わったとき)
            if(changedParameterId == countupModeParameterId)
            {
                // 「選択したバルーンの次から」が選ばれたとき
                if(countupModeSelectedIndex == (int)CountupMode.CountupMode_NextSelectedBalloon)
                {
                    //「モード」パラメータの後に「カウントを始めるバルーン図形」パラメータ追加
                    ParameterItem countupModeParameterItem = command.GetParameterItemByID(countupModeParameterId);
                    int countupModeParameterItemIndex = countupModeParameterItem.Index;

                    command.ParameterGroups.Insert(countupModeParameterItemIndex + 1, ParameterItemType.Shape, startBalloonShapeParameterId, Properties.Resources.StartBalloonShapeParameterName);
                    ShapeParameterItem shapeParameterItem = (ShapeParameterItem)command.GetParameterItemByID(startBalloonShapeParameterId);
                    shapeParameterItem.ShapeTypeFilter = ShapeTypeFilters.Balloon;
                    shapeParameterItem.Description = Properties.Resources.StartBalloonShapeParameterDescription;
                    shapeParameterItem.ValueSaveMode = ParameterValueSaveMode.InCommand;// コマンド実行中は保持
                }
                else
                {
                    //「カウントを始めるバルーン図形」パラメータ削除
                    ParameterItem shapeParameterItem = command.GetParameterItemByID(startBalloonShapeParameterId);
                    if(shapeParameterItem != null)
                    {
                        CommandParameterGroup shapeParameterItemGroup = shapeParameterItem.Group;
                        command.ParameterGroups.Remove(shapeParameterItemGroup);
                    }

                    // 自動にしたとき文字が空だった場合
                   // if(countupModeSelectedIndex == (int)CountupMode.CountupMode_Auto && text == "")
                   //     text = "1";
                }

                // パラメータ値の保持の仕方
                textParameterItem = (TextParameterItem)command.GetParameterItemByID(textParameterId);
                if(textParameterItem != null)
                {
                    if(countupModeSelectedIndex == (int)CountupMode.CountupMode_Manual)
                        textParameterItem.ValueSaveMode = ParameterValueSaveMode.OneTime;
                    else
                        textParameterItem.ValueSaveMode = ParameterValueSaveMode.Always;
                }
            }

            // 「カウントを始めるバルーン図形」パラメーターに変更があった取得
            ShapeParameterItem startBalloonShapeParameterItem = (ShapeParameterItem)command.GetParameterItemByID(startBalloonShapeParameterId);
            if(changedParameterId == startBalloonShapeParameterId)
            {
                if(startBalloonShapeParameterItem != null)
                {
                    string preBalloonText = "";
                    if(!startBalloonShapeParameterItem.IsEmpty)
                    {
                        SelectedShape selectedShape = startBalloonShapeParameterItem.Shape;
                        BalloonShape preBalloon = (BalloonShape)selectedShape.Shape;
                        preBalloonText = preBalloon.Text;
                    }
                    
                    // 次のバルーン文字を「文字」パラメーターにセット
                    SetCountupedTextToParameterItem(command, preBalloonText);
                }
            }

            // カウントを始めるバルーン図形」パラメーター があって空の場合は何もしない
            if(countupModeSelectedIndex == (int)CountupMode.CountupMode_NextSelectedBalloon)
            {
                if(startBalloonShapeParameterItem == null)
                    return;

                if(startBalloonShapeParameterItem.IsEmpty)
                    return;
            }

            // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」パラメーターグループ
            Debug.Assert(locateBalloonModeParameterGroup != null);
            int locateBalloonModeItemID = locateBalloonModeParameterGroup.CurrentItem.ID;

            //ドキュメントの取得
            Document doc = ActiveDocument;

            // 「文字」パラメーターの取得
            string text = "";
            textParameterItem = (TextParameterItem)command.GetParameterItemByID(textParameterId);
            if(textParameterItem != null && !textParameterItem.IsEmpty)
                text = textParameterItem.Text;

            if(text == "")
                return;

            ListParameterItem replaceSameTextBalloonFlgParameterItem = null;

            // 「置き換えるバルーン図形」の場合
            if(locateBalloonModeItemID == replaceBalloonParameterId)
            {
                // 「置き換えるバルーン図形」パラメーターの取得
                ShapeParameterItem replaceBalloonParameterItem = (ShapeParameterItem)command.GetParameterItemByID(replaceBalloonParameterId);
                if(replaceBalloonParameterItem == null)
                    return;

                if(replaceBalloonParameterItem.IsEmpty)
                    return;

                SelectedShape selectedShape = replaceBalloonParameterItem.Shape;
                BalloonShape selectedBalloon = (BalloonShape)selectedShape.Shape;

                // 変更可能な図形のみ選択可能
                Debug.Assert(replaceBalloonParameterItem.EditableShapeOnly);
                // 変更可能な部分図とレイヤの図形か
                //if(!(selectedBalloon.Drawing.Editable && selectedBalloon.Layer.Editable))
                //    return;

                List<Shape> replaceBalloons = new List<Shape>();

                // 「同じ文字のバルーンも置き換える」が「はい」の場合
                replaceSameTextBalloonFlgParameterItem = (ListParameterItem)command.GetParameterItemByID(replaceSameTextBalloonFlgParameterId);
                Debug.Assert(replaceSameTextBalloonFlgParameterItem != null);
                if(replaceSameTextBalloonFlgParameterItem == null)
                    return;

                bool replaceSameTextBalloon = replaceSameTextBalloonFlgParameterItem.SelectedIndex == 0 ? true:false;
                if(replaceSameTextBalloon)
                {
                    // 全部分図から
                    DrawingCollection drawings = doc.Drawings;
                    foreach(Drawing drawing in drawings)
                    {
                        // LINQ の Select、Where を使って
                        // 全部分図の図形から同じ文字のバルーンを取得する
                        // 同時に変更可能な部分図とレイヤの図形に絞り込む
                        ShapeCollection drawingShapes = drawing.Shapes;
                        var balloonShapes = from drawingShape in drawingShapes
                                where drawingShape is BalloonShape where ((BalloonShape)drawingShape).Text == selectedBalloon.Text
                                where drawingShape.Drawing.Editable where drawingShape.Layer.Editable
                                select (BalloonShape)drawingShape;

                        // 置き換えるバルーンリストに追加
                        foreach(BalloonShape balloonShape in balloonShapes)
                            replaceBalloons.Add(balloonShape);
                    }
                }
                else
                {
                    replaceBalloons.Add(selectedBalloon);
                }

                // UNDO 可能な操作を開始
                doc.UndoManager.BeginUndoUnit();

                // バルーン文字の変更
                foreach(BalloonShape balloonShape in replaceBalloons)
                    balloonShape.Text = text;

                // UNDO 可能な操作を終了
                doc.UndoManager.EndUndoUnit();
            }
            else
            {
                Point2d[] pointList = null;

                 // 「始点」の場合
                if(locateBalloonModeItemID == startPointParameterId)
                {
                    // 「始点」パラメーターの取得
                    PointParameterItem pointParameterItem1 = (PointParameterItem)command.GetParameterItemByID(startPointParameterId);
                    if(pointParameterItem1 == null)
                        return;

                    if(pointParameterItem1.IsEmpty)
                        return;

                    // 「終点」パラメーターの取得
                    PointParameterItem pointParameterItem2 = (PointParameterItem)command.GetParameterItemByID(endPointParameterId);
                    if(pointParameterItem2 == null)
                        return;

                    if(pointParameterItem2.IsEmpty)
                        return;

                    pointList = new Point2d[2];
                    pointList[0] = pointParameterItem1.Point;
                    pointList[1] = pointParameterItem2.Point;
                }
                // 「点列」の場合
                else if(locateBalloonModeItemID == pointListParameterId)
                {
                    // 「点列」パラメーターの取得
                    PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetParameterItemByID(pointListParameterId);
                    if(pointListParameterItem == null)
                        return;

                    // 点列の入力が完了していない場合
                    if(pointListParameterItem.IsEmpty || !pointListParameterItem.Complete)
                        return;

                    pointList = pointListParameterItem.Points;
                    if(pointList.Length <= 1)
                        return;
                }
                // 「配置点」の場合
                else if(locateBalloonModeItemID == locatePointParameterId)
                {
                    // 「配置点」パラメーターの取得
                    PointParameterItem pointParameterItem1 = (PointParameterItem)command.GetParameterItemByID(locatePointParameterId);
                    if(pointParameterItem1 == null)
                        return;

                    if(pointParameterItem1.IsEmpty)
                        return;

                    pointList = new Point2d[1];
                    pointList[0] = pointParameterItem1.Point;
                }

                // 図形作成
                List<Shape> shapes = null;
                int stat = CreateCommandShapes(pointList, text, out shapes);
                if(stat < 0)
                {
                    //ShowErrMessage(stat);

                    e.CommandState = CommandState.ParameterError;
                    return;
                }

                // UNDO 可能な操作を開始
                doc.UndoManager.BeginUndoUnit();

                Drawing currentDrawing = doc.CurrentDrawing;

                // 図面に図形追加
                foreach(Shape shape in shapes)
                {
                    currentDrawing.Shapes.Add(shape);
                }

                // UNDO 可能な操作を終了
                doc.UndoManager.EndUndoUnit();
            }

            // コマンド実行
            e.CommandState = CommandState.Execute;

            if(textParameterItem != null)
            {
                // 「自動」と「選択したバルーンの次から」の場合
                if(countupModeSelectedIndex == (int)CountupMode.CountupMode_Auto
                    || countupModeSelectedIndex == (int)CountupMode.CountupMode_NextSelectedBalloon)
                {
                    // 「文字」パラメーターに次の文字をセット
                    string oldText = textParameterItem.Text;
                    SetCountupedTextToParameterItem(command, oldText);

                    // 「文字」パラメータの値を設定に保持
                    addInSettings.textParameterValue = textParameterItem.Text;
                }
            }

            // 「モード」パラメーターのインデックス保持
            addInSettings.countupModeListParameterItemIndex = countupModeListParameterItem.SelectedIndex;

            // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」の選択アイテムID保持
            addInSettings.locateBalloonModeItemID = locateBalloonModeItemID;

            // 同じ文字のバルーンも置き換えるかのフラグインデックス 保持
            if(replaceSameTextBalloonFlgParameterItem != null)
            {
                addInSettings.replaceSameTextBalloonFlgIndex = replaceSameTextBalloonFlgParameterItem.SelectedIndex;
            }
        }

        /// <summary>
        /// マウス移動イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Command_MouseMove(Object sender, CommandMouseMoveEventArgs e)
        {
            Command command = (Command)sender;
            // Command.GetTemporaryParameterItemByID を使用してコマンドの一時的なパラメータアイテムを取得します。
            // 一時的なパラメータアイテムからは、すでにパラメータの入力が確定している場合は確定している入力値を取得できます。
            // 入力が確定していない場合はマウスの現在位置などの一時的な値が取得できます。
            // Gets a command's temporary parameter item by Command.GetTemporaryParameterItemByID.
            // From a temporary parameter item, you can get the committed input value if the input of the parameter has already been committed.
            // If the input is not finalized, you can get temporary values such as the current position of the mouse.

            // 「モード」のパラメータアイテム取得
            ListParameterItem countupModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(countupModeParameterId);
            if(countupModeListParameterItem == null)
                return;
            // 選ばれているモードのインデックス
            int countupModeSelectedIndex = countupModeListParameterItem.SelectedIndex;

            // 「カウントを始めるバルーン図形」パラメーター があって空の場合は何もしない
            if(countupModeSelectedIndex == (int)CountupMode.CountupMode_NextSelectedBalloon)
            {
                ShapeParameterItem startBalloonShapeParameterItem = (ShapeParameterItem)command.GetParameterItemByID(startBalloonShapeParameterId);
                if(startBalloonShapeParameterItem == null)
                    return;

                if(startBalloonShapeParameterItem.IsEmpty)
                    return;
            }

            // 「始点」・「点列」・「配置点」・「置き換えるバルーン図形」パラメーターグループ
            Debug.Assert(locateBalloonModeParameterGroup != null);
            int locateBalloonModeItemID = locateBalloonModeParameterGroup.CurrentItem.ID;

            // 「文字」パラメーターの取得
            string text = "";
            TextParameterItem testParameterItem = (TextParameterItem)command.GetTemporaryParameterItemByID(textParameterId);
            if(testParameterItem != null)
                text = testParameterItem.Text;


            Point2d[] pointList = null;

            // 「置き換えるバルーン図形」の場合
            if(locateBalloonModeItemID == replaceBalloonParameterId)
            {
                // 「置き換えるバルーン図形」パラメーターの取得
                ShapeParameterItem replaceBalloonParameterItem = (ShapeParameterItem)command.GetTemporaryParameterItemByID(replaceBalloonParameterId);
                if(replaceBalloonParameterItem == null)
                    return;

                if(replaceBalloonParameterItem.IsEmpty)
                    return;

                SelectedShape selectedShape = replaceBalloonParameterItem.Shape;
                BalloonShape selectedBalloon = (BalloonShape)selectedShape.Shape;

                // 変更可能な図形のみ選択可能
                Debug.Assert(replaceBalloonParameterItem.EditableShapeOnly);

                Point2d[] points = selectedBalloon.Points;
                int pointsCount = points.Length;

                pointList = new Point2d[1];
                pointList[0] = points[pointsCount-1];

            }
            else
            {
                // 「始点」の場合
                if(locateBalloonModeItemID == startPointParameterId)
                {
                    // 「始点」パラメーターの取得
                    PointParameterItem pointParameterItem1 = (PointParameterItem)command.GetTemporaryParameterItemByID(startPointParameterId);
                    if(pointParameterItem1 == null)
                        return;

                    // 「終点」パラメーターの取得
                    PointParameterItem pointParameterItem2 = (PointParameterItem)command.GetTemporaryParameterItemByID(endPointParameterId);
                    if(pointParameterItem2 == null)
                        return;

                    // 同一点の場合
                    if(Geometry.IsPointEquals(pointParameterItem1.Point, pointParameterItem2.Point))
                        return;

                    pointList = new Point2d[2];
                    pointList[0] = pointParameterItem1.Point;
                    pointList[1] = pointParameterItem2.Point;
                }
                // 「点列」の場合
                else if(locateBalloonModeItemID == pointListParameterId)
                {
                    // 「点列」パラメーターの取得
                    PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetTemporaryParameterItemByID(pointListParameterId);
                    if(pointListParameterItem == null)
                        return;

                    pointList = pointListParameterItem.Points;
                    if(pointList.Length <= 1)
                        return;
                }
                // 「配置点」の場合
                else if(locateBalloonModeItemID == locatePointParameterId)
                {
                    // 「配置点」パラメーターの取得
                    PointParameterItem pointParameterItem1 = (PointParameterItem)command.GetTemporaryParameterItemByID(locatePointParameterId);
                    if(pointParameterItem1 == null)
                        return;

                    pointList = new Point2d[1];
                    pointList[0] = pointParameterItem1.Point;
                }
            }
            
            // 図形作成
            List<Shape> shapes = null;
            int stat = CreateCommandShapes(pointList, text, out shapes);
            if(stat != RC_ERR_SUCCESS)
                return;

            // ラバーバンド図形リストに追加
            List<Shape> rubberBandShapes = new List<Shape>();
            for(int i = 0; i < shapes.Count; i++)
                rubberBandShapes.Add(shapes[i]);

            // ラバーバンド図形描画
            if(rubberBandShapes.Count > 0)
                command.DrawShapeRubberBand(rubberBandShapes.ToArray());

        }

        /// <summary>
        /// 出力図形の作成
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="text"></param>
        /// <param name="shapes"></param>
        /// <returns></returns>
        private int CreateCommandShapes(Point2d[] pointList, string text, out List<Shape> shapes)
        {
            shapes = null;

            if(pointList.Length < 1)
                return RC_ERR_NONPOINTS;

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;
            // 書込みレイヤ
            Layer currentLayer = doc.LayerTable.CurrentLayer;

            // バルーン図形を作成
            BalloonShape balloonShape = ShapeFactory.CreateBalloon(text, pointList);

            doc.BalloonSettings.CopyToShape(balloonShape);// 既定値をセット
            balloonShape.Layer = currentLayer;// レイヤ

            // フィレットされた結果の図形
            shapes = new List<Shape>();// 作成図形
            shapes.Add(balloonShape);
            
            return RC_ERR_SUCCESS;
        }


        /// <summary>
        /// 「文字」パラメーターに次の文字をセット
        /// </summary>
        /// <param name="text"></param>
        private void SetCountupedTextToParameterItem(Command command, string text)
        {
            TextParameterItem textParameterItem = (TextParameterItem)command.GetParameterItemByID(textParameterId);
            if(textParameterItem != null)
            {
                // 「文字」パラメーターを次の文字に変更
                changedTextParameter = true;// パラメーターの内容を変更するとイベントが発生するのでその時は何もしないようにするためのフラグ
                textParameterItem.Text = GetNextBalloonText(text);// 次の文字をセット
            }
        }

        /// <summary>
        /// 次のバルーン文字取得
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        private string GetNextBalloonText(string text)
        {
            string nextText = "";
 
            if(text != "")
            {
                int number = 0;

                // 数字の場合(数字は半角のみ対応)
                if(int.TryParse(text, out number))
                {
                    number++;// Countup
                    nextText = number.ToString();
                }
                // 1文字の場合
                else if(text.Length == 1)
                {
                    // 定義済みの文字リストから同じ文字を探して
                    // あれば次の文字をセット
                    int listLength = balloonTextList.Length;
                    for(int j = 0; j < listLength; j++)
                    {
                        int index = -1;
                        string listString = balloonTextList[j].ToString();

                        if((index = listString.IndexOf(text)) >= 0)
                        {
                            int length = listString.Length;
                            if(index == length - 1)
                                nextText = listString[0].ToString();
                            else
                                nextText = listString[index + 1].ToString();
                        }
                    }
                }
                // 2文字以上の場合
                else
                {
                    // 2段のバルーン文字の場合 ex. 1/A, a/10
                    int index = -1;
                    string returnText = "\\|";
                    if((index = text.IndexOf(returnText)) >= 0)
                    {
                        string upperText = text.Substring(0, index);// 上段の文字
                        string lowerText = text.Substring(index + returnText.Length);// 下段の文字

                        // 上段が数字の場合
                        if(int.TryParse(upperText, out number))
                        {
                            number++;// Countup
                            nextText = number.ToString() + returnText + lowerText;
                        }
                        // 下段が数字の場合
                        else if(int.TryParse(lowerText, out number))
                        {
                            number++;// Countup
                            nextText = upperText + returnText + number.ToString();
                        }
                    }
                    // 先頭が希望文字でその後ろが数字の場合はカウントアップさせる
                    // 3文字以上で2文字目が"-"の場合 ex.A-1, あ-1
                    else if(text.Length > 2 && text[1].ToString() == "-")
                    {
                        string rightText = text.Substring(2);
                        if(int.TryParse(rightText, out number))
                        {
                            number++;// Countup
                            nextText = text.Substring(0, 2) + number.ToString();// A-1 -> A-2
                        }
                    }
                    // 2文字以上の場合 ex. A1, あ1
                    else if(text.Length > 1)
                    {
                        string rightText = text.Substring(1);
                        if(int.TryParse(rightText, out number))
                        {
                            number++;// Countup
                            nextText = text.Substring(0, 1) + number.ToString();// A1 -> A2
                        }
                    }
                }
            }

            return nextText;
        }

    }
}
